import logging
import json
import click

from quart import Quart, request, render_template, redirect

from .data_provider import DataProvider
from .types import Message
from .server import BlueprintForActors
from .server.verify_actor import ActorVerifier

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@click.command()
@click.option("--only_generate_config", default=False, is_flag=True)
@click.option("--port", default=2909, help="port to run on")
@click.option(
    "--domain", default="localhost:2909", help="domain the service is running on"
)
def verify_actor(only_generate_config, port, domain):
    dp = DataProvider.generate_and_load(only_generate_config)
    app = Quart(__name__)

    actor_list = dp.possible_actors + [dp.one_actor]
    app.register_blueprint(BlueprintForActors(actor_list).blueprint)

    @app.get("/")
    async def index():
        return await render_template("index.html.j2")

    @app.post("/verify")
    async def verify():
        form_data = await request.form
        actor_uri = form_data.get("actor_uri")
        if not actor_uri:
            return redirect("/")

        message = Message()
        message.add(f"Got Actor Uri {actor_uri}")

        verifier = ActorVerifier(
            actor_list=actor_list, remote_uri=actor_uri, message=message, domain=domain
        )

        result = await verifier.verify()

        if "json" in request.headers.get("accept"):
            return result

        return await render_template(
            "verify_actor_result.html.j2",
            messages=json.dumps(message.response, indent=2),
            result=result,
            actor_uri=actor_uri,
        )

    app.run(port=port, host="0.0.0.0", use_reloader=True)


if __name__ == "__main__":
    verify_actor()



import base64
import aiohttp
from typing import Tuple

from bovine.jsonld import with_external_context
from bovine.utils.date import parse_gmt, check_max_offset_now


def find_with_item(key_list, key_id):
    for key in key_list:
        if key.get("id") == key_id:
            return key
    return None


def public_key_owner_from_dict(
    actor: dict, key_id: str
) -> Tuple[str | None, str | None]:
    public_key_data = actor.get("publicKey", {})

    if isinstance(public_key_data, list):
        if len(public_key_data) == 1:
            public_key_data = public_key_data[0]
        else:
            public_key_data = find_with_item(public_key_data, key_id)

    if not public_key_data:
        return None, None

    public_key = public_key_data.get("publicKeyPem")
    owner = public_key_data.get("owner")

    return public_key, owner


def actor_object_to_public_key(
    actor: dict, key_id: str
) -> Tuple[str | None, str | None]:
    public_key, owner = public_key_owner_from_dict(with_external_context(actor), key_id)

    if public_key and owner:
        return public_key, owner

    return public_key_owner_from_dict(actor, key_id)


def validate_basic_signature_fields(message, request, signature):
    if "(request-target)" not in signature.fields:
        return message.error("(request-target) must be a signature field")

    if "host" not in signature.fields:
        return message.error("host must be a signature field")

    if "date" not in signature.fields:
        return message.error("date must be a signature field")

    date_header = request.headers.get("date")
    message.add(f"Got date header {date_header}")

    try:
        date_parsed = parse_gmt(date_header)
    except Exception as e:
        return message.error(["Failed to parse date", repr(e)])

    if not check_max_offset_now(date_parsed):
        return message.error(["Date not within the last 5 minutes"])

    return None


async def retrieve_public_key(message, key_id, request):
    if key_id == "about:inline":
        try:
            public_key = base64.standard_b64decode(
                request.headers.get("X-Public-Key")
            ).decode("utf-8")
        except Exception:
            public_key = None
        if not public_key:
            return None, message.error(
                "Please set the header 'X-Public-Key' to provide a public key"
            )
    else:
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(key_id) as response:
                    if "json" in response.headers.get("content-type"):
                        data = await response.json()
                        public_key, _ = actor_object_to_public_key(data)
                    else:
                        public_key = await response.text()
        except Exception as e:
            return (
                None,
                message.error(
                    [
                        "Failed to fetch public key",
                        "Use about:inline to include a public key in a header",
                        "    about:inline is expected to be base64 encoded",
                        "or make sure keyId can be resolved to a public key",
                        repr(e),
                    ]
                ),
            )

    return public_key, None



import aiohttp
import logging
import secrets
import json
from typing import List

from dataclasses import dataclass

from bovine.clients import lookup_uri_with_webfinger

from fediverse_pasture.data_provider import bovine_actor_for_actor_data
from fediverse_pasture.data_provider.models import ActorData
from fediverse_pasture.types import Message

logger = logging.getLogger(__name__)


def can_be_resolved(uri):
    if uri.startswith("http://"):
        return True

    if uri.startswith("https://"):
        return True

    return False


@dataclass
class ActorVerifier:
    """Class implementing the logic to verify a remote actor"""

    actor_list: List[ActorData]
    remote_uri: str
    domain: str
    message: Message

    remote_actor_uri: str | None = None

    async def verify(self):
        """Public interface, this method loops over the actors in `actor_list`
        and verifies using them if the remote_uri is accessible.

        Returns a dictionary with actor.actor_name as key and the
        result as value"""
        async with aiohttp.ClientSession() as session:
            await self.determine_remote_actor_uri(session)
            return {
                actor.actor_name: await self.verify_for_actor(session, actor)
                for actor in self.actor_list
                if actor.actor_name != "actor"
            }

    async def verify_for_actor(self, session, actor):
        actor_id = f"http://{self.domain}/{actor.actor_name}"
        bovine_actor, _ = bovine_actor_for_actor_data(actor_id, actor)
        await bovine_actor.init(session=session)

        self.message.add(f"Running verification for {actor.actor_name}")

        if actor.requires_signed_post_for_inbox:
            return await self.fetch_remote_and_post_using_actor(bovine_actor)
        else:
            return await self.fetch_remote_and_post_using_session(session)

    async def fetch_remote_and_post_using_actor(self, bovine_actor):
        result = {"get_actor": False, "post_inbox": False}

        try:
            actor = await bovine_actor.get(self.remote_actor_uri)
            inbox = actor.get("inbox")

            self.message.add(f"Got inbox {inbox}")

            if inbox:
                result["get_actor"] = True
                try:
                    response = await bovine_actor.post(
                        inbox,
                        data={
                            "@context": "https://www.w3.org/ns/activitystreams",
                            "type": "Like",
                            "actor": bovine_actor.actor_id,
                            "id": bovine_actor.actor_id + secrets.token_urlsafe(8),
                        },
                    )
                    self.message.add("Successfully posted to inbox with result")
                    self.message.add(response.status)
                    self.message.add((await response.text()))
                    result["post_inbox"] = True
                except Exception as e:
                    self.message.add("Failed to post to inbox")
                    self.message.add(repr(e))

        except Exception as e:
            self.message.add("Something went wrong")
            self.message.add(repr(e))

        return result

    async def fetch_remote_and_post_using_session(self, session):
        result = {"get_actor": False, "post_inbox": False}

        try:
            async with session.get(
                self.remote_actor_uri, headers={"accept": "application/activity+json"}
            ) as response:
                actor = json.loads(await response.text())
        except Exception as e:
            self.message.add("Something went wrong when fetching actor")
            self.message.add(repr(e))
            return result

        inbox = actor.get("inbox")

        self.message.add(f"Got inbox {inbox}")

        if inbox:
            result["get_actor"] = True
            try:
                async with session.post(
                    inbox,
                    data={
                        "@context": "https://www.w3.org/ns/activitystreams",
                        "type": "EchoRequest",
                    },
                ) as response:
                    self.message.add(f"Got {response.status} for unsigned POST")

                    if response.status < 400 and response.status > 100:
                        result["post_inbox"] = True
            except Exception as e:
                self.message.add("Something went wrong when posting to inbox")
                self.message.add(repr(e))
                return result

        return result

    async def determine_remote_actor_uri(self, session) -> None:
        """Used to resolve an acct-URI into the corresponding http/https-URI
        using webfinger"""
        if can_be_resolved(self.remote_uri):
            self.remote_actor_uri = self.remote_uri
            self.message.add(f"Can fetch actor from {self.remote_uri}")
            return

        self.message.add(f"Need to resolve {self.remote_uri} to actor object id")

        if self.remote_uri.startswith("acct:"):
            acct_uri = self.remote_uri
        else:
            self.message.add("Not in account uri format")

            if self.remote_uri[0] == "@":
                acct_uri = "acct:" + self.remote_uri[1:]
            else:
                acct_uri = "acct:" + self.remote_uri

        domain = acct_uri.split("@")[1]

        self.message.add(f"Resolving {acct_uri} using webfinger")

        try:
            object_id, _ = await lookup_uri_with_webfinger(
                session, acct_uri, f"http://{domain}"
            )
        except Exception as e:
            logger.exception(e)
            return

        self.message.add(f"Resolved to {object_id}")
        self.remote_actor_uri = object_id



import base64
import logging
import click
from quart import Quart, request

from bovine.crypto.helper import sign_message
from bovine.crypto.signature import Signature
from bovine.crypto.http_signature import build_signature

from .server.utils import validate_basic_signature_fields, retrieve_public_key
from .types import Message

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@click.command()
@click.option("--port", default=2909, help="port to run on")
def http_signature(port):
    app = Quart(__name__)

    @app.get("/")
    async def get_handler():
        message = Message()
        message.add("Got get request")

        signature_header = request.headers.get("signature")
        if not signature_header:
            return message.error("Signature header is missing"), 401

        message.add(f"Signature header '{signature_header}'")

        try:
            signature = Signature.from_signature_header(signature_header)
        except Exception as e:
            return (
                message.error(
                    [
                        "Failed to parse signature",
                        repr(e),
                    ]
                ),
                401,
            )

        message.add(f"""Got fields {", ".join(signature.fields)}""")

        error_dict = validate_basic_signature_fields(message, request, signature)

        if error_dict:
            return error_dict, 401

        header_fields = [x for x in signature.fields if x[0] != "("]

        http_signature = build_signature("http_signature", "get", request.path)

        for x in header_fields:
            if x != "host":
                http_signature.with_field(x, request.headers.get(x))

        message.add(f"""Message to sign "{http_signature.build_message()}" """)
        message.add(f"Got key id {signature.key_id}")

        public_key, error_dict = await retrieve_public_key(
            message, signature.key_id, request
        )

        if error_dict:
            return error_dict, 401

        message.add(f"""Got public key "{public_key}" """)

        try:
            result = http_signature.verify(
                public_key=public_key, signature=signature.signature
            )
        except Exception as e:
            return (
                message.error(
                    [
                        "Something went wrong when verifying signature",
                        repr(e),
                    ]
                ),
                401,
            )

        if not result:
            private_key = request.headers.get("X-Private-Key")

            if private_key is None:
                message.add(
                    "Set X-Private-Key to your base64 encoded private key\
                        for expected signature",
                )
            else:
                message.add("Expected signature:")
                message.add(
                    sign_message(
                        private_key=base64.standard_b64decode(private_key).decode(
                            "utf-8"
                        ),
                        message=http_signature.build_message(),
                    )
                )

            return message.error("Invalid signature"), 401
        else:
            message.add("SUCCESS!!!")

        return message.response, 200 if result else 401

    app.run(port=port, host="0.0.0.0", use_reloader=True)


if __name__ == "__main__":
    http_signature()



import aiohttp
from dataclasses import dataclass
from datetime import datetime
import bovine

from fediverse_pasture.types import ApplicationAdapterForLastActivity


@dataclass
class FirefishApplication:
    domain: str
    username: str
    session: aiohttp.ClientSession | None = None

    async def determine_actor_uri(self):
        actor_uri, _ = await bovine.clients.lookup_uri_with_webfinger(
            self.session, f"acct:{self.username}@{self.domain}", f"http://{self.domain}"
        )
        return actor_uri

    async def top_public(self):
        response = await self.session.post(
            f"http://{self.domain}/api/notes/global-timeline"
        )
        public_timeline = await response.json()
        return public_timeline[0]

    async def top_public_with_published(self, published: datetime) -> dict | None:
        data = await self.top_public()
        created_at = data.get("createdAt")
        created_at = datetime.fromisoformat(created_at.removesuffix("Z"))
        if created_at == published:
            return data
        return None

    async def last_activity(self, session: aiohttp.ClientSession):
        self.session = session

        actor_uri = await self.determine_actor_uri()

        return ApplicationAdapterForLastActivity(
            actor_uri=actor_uri,
            fetch_activity=self.top_public_with_published,
            application_name="firefish",
        )



import aiohttp
from .mastodon import MastodonApplication
from .firefish import FirefishApplication


async def activity_for_mastodon(
    domain: str, username: str, access_token: str, session: aiohttp.ClientSession
):
    """Creates a ApplicationAdapterForLastActivity object for connecting to
    mastodon. Example usage:

    .. code-block:: python

        mastodon = await activity_for_mastodon("mastodon_web", "bob", "xxx", session)
    """
    mastodon = MastodonApplication(
        domain=domain, access_token=access_token, username=username
    )

    return mastodon.last_activity(session)


async def activity_for_firefish(
    domain: str, username: str, session: aiohttp.ClientSession
):
    """Creates a ApplicationAdapterForLastActivity object for connecting to
    firefish. Example usage:

    .. code-block:: python

        firefish = await activity_for_firefish("firefish_web", "admin", session)
    """
    firefish = FirefishApplication(domain=domain, username=username)

    return await firefish.last_activity(session)



from pydantic import BaseModel
from typing import List


class ActorKeyPair(BaseModel):
    name: str
    public: str
    private: str


class ActorData(BaseModel):
    actor_name: str
    key_pairs: List[ActorKeyPair] = []
    user_part: str | None = None

    summary: str = ""

    requires_signed_get_for_actor: bool = False
    requires_signed_post_for_inbox: bool = False
